package generated.ageVerification.javacard.ePa;

import javacard.framework.ISO7816;
import javacard.framework.ISOException;
import javacard.framework.Util;
/**
 * Class for T(L)V encoding of messages and other classes
 * 
 */
public class Coding {

	/**
	 * offset into encodeDestination for writing/reading
	 */
	private short encoding_length;

	/**
	 * array used for encoding/decoding
	 */
	private byte[] encodeDestination;

	/**
	 * Singleton instance
	 */
	private static Coding the_instance;

	/**
	 * Singleton accessor
	 * 
	 * @return Singleton instance
	 */

	public static Coding getInstance() {
		if (the_instance == null)
			the_instance = new Coding();
		return the_instance;
	}
	/**
	 * return offset into encodeDestination
	 * 
	 * @return value of encoding_length
	 */
	public short getEncodingLength() {
		return encoding_length;
	}

	/**
	 * initializes encoding. This method must be called before encoding an object.
	 * The offset (encoding_length) is reset to 0 and encodeDestination is set to dest.
	 * 
	 * @param dest The (empty) destination array for encoding
	 */
	public void encodeInit(byte[] dest) {
		encodeDestination = dest;
		encoding_length = 0;
	}
	/**
	 * initializes decoding. The offset (encoding_length) is reset to 0.
	 */
	public void decodeInit() {
		encoding_length = 0;
	}

	/**
	 * initializes encoding and encodes an object.
	 * 
	 * @param o The object to encode
	 * @param dest destination array for the encoding
	 * @return length of the encoded object
	 */

	private byte decodeByte(byte[] in) {
		if (in[encoding_length++] != Code.BYTE) {

			SimpleComm.raiseError();
		}

		byte res = in[encoding_length++];
		return res;
	}

	private boolean decodeBoolean(byte[] in) {
		if (in[encoding_length++] != Code.BOOLEAN) {

			SimpleComm.raiseError();

		}
		boolean res = (in[encoding_length++] == 1 ? true : false);
		return res;
	}

	public short encode(Codeable o, byte[] dest) {
		encodeInit(dest);

		try {
			encode(o);
		}

		catch (ArrayIndexOutOfBoundsException e) {
			SimpleComm.raiseError();
		}
		return encoding_length;

	}

	/**
	 * computes the encoding length of an object without actually encoding it.
	 * 
	 * @param o The object of which the encoding length should be computed
	 * @return encoding length
	 */

	public short computeSize(Codeable o) {
		encoding_length = 0;
		try {
			compSize(o);
			return encoding_length;
		} catch (Exception e) {
			return -1;
		}
	}

	private void encodeShort(short o) {
		encodeDestination[encoding_length++] = Code.SHORT;
		Util.setShort(encodeDestination, encoding_length, o);
		encoding_length += 2;
	}

	public void encodeByte(byte o) {
		encodeDestination[encoding_length++] = Code.BYTE;
		encodeDestination[encoding_length++] = o;
	}

	private void encodeBoolean(boolean o) {
		encodeDestination[encoding_length++] = Code.BOOLEAN;
		encodeDestination[encoding_length++] = (byte) (o ? 1 : 0);
	}

	private void compSizeShort(short o) {
		encoding_length += 3;
	}

	private void compSizeByte(byte o) {
		encoding_length += 2;
	}

	private void compSizeBoolean(boolean o) {
		encoding_length += 2;
	}

	private short decodeShort(byte[] in) {
		if (in[encoding_length++] != Code.SHORT)
			SimpleComm.raiseError();
		short res = Util.getShort(in, encoding_length);
		encoding_length += 2;
		return res;
	}

	public void encodeByteArray(byte[] o) {

		encodeDestination[encoding_length++] = Code.BYTEARRAY;

		Util.setShort(encodeDestination, encoding_length, (short) (o.length));
		encoding_length += 2;

		Util.arrayCopy(o, (short) 0, encodeDestination, encoding_length,
				(short) o.length);

		encoding_length += o.length;
	}

	private void compSizeByteArray(byte[] o) {
		encoding_length += 3;
		encoding_length += o.length;
	}

	private void decodeByteArrayInto(byte[] in, byte[] into) {
		if (in[encoding_length++] != Code.BYTEARRAY)
			SimpleComm.raiseError();
		short len = Util.getShort(in, encoding_length);
		encoding_length += 2;
		Util.arrayCopy(in, encoding_length, into, (short) 0, len);
		encoding_length += len;
	}

	/**
	 * encodes an object without initializing destination and offset.
	 * 
	 * @param c object to encode
	 */

	public void encode(Codeable c) {

		switch (c.getCode()) {
			case Code.SENDCERT :
				encodeSendCert((sendCert) c);
				break;
			case Code.SENDNONCE :
				encodeSendNonce((sendNonce) c);
				break;
			case Code.SIGBACK :
				encodeSigBack((sigBack) c);
				break;
			case Code.SIG :
				encodeSig((sig) c);
				break;
			case Code.NONCE :
				encodeNonce((Nonce) c);
				break;
			case Code.CERTIFICATE :
				encodeCertificate((Certificate) c);
				break;
			case Code.SIGBACKDATA :
				encodeSigBackData((SigBackData) c);
				break;
			case Code.SIGNEDDATA :
				encodeSignedData((SignedData) c);
				break;
			case Code.SIGDATA :
				encodeSigData((SigData) c);
				break;
			case Code.CERTDATA :
				encodeCertData((CertData) c);
				break;
			case Code.DATE :
				encodeDate((Date) c);
				break;
			case Code.PUBLICKEY :
				encodePublicKey((PublicKey) c);
				break;
			case Code.PRIVATEKEY :
				encodePrivateKey((PrivateKey) c);
				break;
			// never reached
			default :

				SimpleComm.raiseError();

		}
	}

	public Message decodeMessage(byte[] in, short offset, short expectedLength) {
		encoding_length = offset;
		Message m = null;
		try {
			switch (in[encoding_length]) {
				case Code.SENDCERT :
					m = Store.newsendCert();
					decodeSendCertInto(in, (sendCert) m);
					break;
				case Code.SENDNONCE :
					m = Store.newsendNonce();
					decodeSendNonceInto(in, (sendNonce) m);
					break;
				case Code.SIGBACK :
					m = Store.newsigBack();
					decodeSigBackInto(in, (sigBack) m);
					break;
				case Code.SIG :
					m = Store.newsig();
					decodeSigInto(in, (sig) m);
					break;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			SimpleComm.raiseError();
		}
		if (m == null || encoding_length != (short) (expectedLength + offset))
			SimpleComm.raiseError();
		return m;
	}

	public void compSize(Codeable c) {
		switch (c.getCode()) {
			case Code.SENDCERT :

				compSizeSendCert((sendCert) c);
				break;
			case Code.SENDNONCE :

				compSizeSendNonce((sendNonce) c);
				break;
			case Code.SIGBACK :

				compSizeSigBack((sigBack) c);
				break;
			case Code.SIG :

				compSizeSig((sig) c);
				break;
			case Code.NONCE :

				compSizeNonce((Nonce) c);
				break;
			case Code.CERTIFICATE :

				compSizeCertificate((Certificate) c);
				break;
			case Code.SIGBACKDATA :

				compSizeSigBackData((SigBackData) c);
				break;
			case Code.SIGNEDDATA :

				compSizeSignedData((SignedData) c);
				break;
			case Code.SIGDATA :

				compSizeSigData((SigData) c);
				break;
			case Code.CERTDATA :

				compSizeCertData((CertData) c);
				break;
			case Code.DATE :

				compSizeDate((Date) c);
				break;
			case Code.PUBLICKEY :

				compSizePublicKey((PublicKey) c);
				break;
			case Code.PRIVATEKEY :

				compSizePrivateKey((PrivateKey) c);
				break;
			// never reached
			default :

				SimpleComm.raiseError();
				break;

		}
	}

	private void encodeSendCert(sendCert c) {
		encodeDestination[encoding_length++] = Code.SENDCERT;
		encode(c.Nonce1);
		encode(c.TCert);
	}

	private void decodeSendCertInto(byte[] in, sendCert into) {
		if (in[encoding_length++] != Code.SENDCERT)
			SimpleComm.raiseError();
		decodeNonceInto(in, (Nonce) (into.Nonce1));
		decodeCertificateInto(in, (Certificate) (into.TCert));
	}

	private void compSizeSendCert(sendCert c) {
		encoding_length++;
		compSize(c.Nonce1);
		compSize(c.TCert);
	}

	private void encodeSendNonce(sendNonce c) {
		encodeDestination[encoding_length++] = Code.SENDNONCE;
		encode(c.Nonce2);
	}

	private void decodeSendNonceInto(byte[] in, sendNonce into) {
		if (in[encoding_length++] != Code.SENDNONCE)
			SimpleComm.raiseError();
		decodeNonceInto(in, (Nonce) (into.Nonce2));
	}

	private void compSizeSendNonce(sendNonce c) {
		encoding_length++;
		compSize(c.Nonce2);
	}

	private void encodeSigBack(sigBack c) {
		encodeDestination[encoding_length++] = Code.SIGBACK;
		encode(c.data);
		encode(c.sig);
		encode(c.EPaCert);
	}

	private void decodeSigBackInto(byte[] in, sigBack into) {
		if (in[encoding_length++] != Code.SIGBACK)
			SimpleComm.raiseError();
		decodeSigBackDataInto(in, (SigBackData) (into.data));
		decodeSignedDataInto(in, (SignedData) (into.sig));
		decodeCertificateInto(in, (Certificate) (into.EPaCert));
	}

	private void compSizeSigBack(sigBack c) {
		encoding_length++;
		compSize(c.data);
		compSize(c.sig);
		compSize(c.EPaCert);
	}

	private void encodeSig(sig c) {
		encodeDestination[encoding_length++] = Code.SIG;
		encode(c.data);
		encode(c.sig);
	}

	private void decodeSigInto(byte[] in, sig into) {
		if (in[encoding_length++] != Code.SIG)
			SimpleComm.raiseError();
		decodeSigDataInto(in, (SigData) (into.data));
		decodeSignedDataInto(in, (SignedData) (into.sig));
	}

	private void compSizeSig(sig c) {
		encoding_length++;
		compSize(c.data);
		compSize(c.sig);
	}

	private void encodeNonce(Nonce c) {
		encodeDestination[encoding_length++] = Code.NONCE;
		encodeByteArray((byte[]) c.nonce);
	}

	private void decodeNonceInto(byte[] in, Nonce into) {
		if (in[encoding_length++] != Code.NONCE)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.nonce);
	}

	private void compSizeNonce(Nonce c) {
		encoding_length++;
		compSizeByteArray(c.nonce);
	}

	private void encodeCertificate(Certificate c) {
		encodeDestination[encoding_length++] = Code.CERTIFICATE;
		encode(c.data);
		encode(c.sig);
	}

	private void decodeCertificateInto(byte[] in, Certificate into) {
		if (in[encoding_length++] != Code.CERTIFICATE)
			SimpleComm.raiseError();
		decodeCertDataInto(in, (CertData) (into.data));
		decodeSignedDataInto(in, (SignedData) (into.sig));
	}

	private void compSizeCertificate(Certificate c) {
		encoding_length++;
		compSize(c.data);
		compSize(c.sig);
	}

	private void encodeSigBackData(SigBackData c) {
		encodeDestination[encoding_length++] = Code.SIGBACKDATA;
		encode(c.Nonce1);
		encode(c.Nonce2);
		encodeShort((short) c.age);
		encodeByteArray((byte[]) c.answer);
	}

	private void decodeSigBackDataInto(byte[] in, SigBackData into) {
		if (in[encoding_length++] != Code.SIGBACKDATA)
			SimpleComm.raiseError();
		decodeNonceInto(in, (Nonce) (into.Nonce1));
		decodeNonceInto(in, (Nonce) (into.Nonce2));
		into.age = decodeShort(in);
		decodeByteArrayInto(in, into.answer);
	}

	private void compSizeSigBackData(SigBackData c) {
		encoding_length++;
		compSize(c.Nonce1);
		compSize(c.Nonce2);
		compSizeShort(c.age);
		compSizeByteArray(c.answer);
	}

	private void encodeSignedData(SignedData c) {
		encodeDestination[encoding_length++] = Code.SIGNEDDATA;
		encodeByteArray((byte[]) c.signed);
	}

	private void decodeSignedDataInto(byte[] in, SignedData into) {
		if (in[encoding_length++] != Code.SIGNEDDATA)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.signed);
	}

	private void compSizeSignedData(SignedData c) {
		encoding_length++;
		compSizeByteArray(c.signed);
	}

	private void encodeSigData(SigData c) {
		encodeDestination[encoding_length++] = Code.SIGDATA;
		encode(c.Nonce1);
		encode(c.Nonce2);
		encodeShort((short) c.age);
		encode(c.Datum);
	}

	private void decodeSigDataInto(byte[] in, SigData into) {
		if (in[encoding_length++] != Code.SIGDATA)
			SimpleComm.raiseError();
		decodeNonceInto(in, (Nonce) (into.Nonce1));
		decodeNonceInto(in, (Nonce) (into.Nonce2));
		into.age = decodeShort(in);
		decodeDateInto(in, (Date) (into.Datum));
	}

	private void compSizeSigData(SigData c) {
		encoding_length++;
		compSize(c.Nonce1);
		compSize(c.Nonce2);
		compSizeShort(c.age);
		compSize(c.Datum);
	}

	private void encodeCertData(CertData c) {
		encodeDestination[encoding_length++] = Code.CERTDATA;
		encodeByteArray((byte[]) c.Name);
		encode(c.PublicKey);
	}

	private void decodeCertDataInto(byte[] in, CertData into) {
		if (in[encoding_length++] != Code.CERTDATA)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.Name);
		decodePublicKeyInto(in, (PublicKey) (into.PublicKey));
	}

	private void compSizeCertData(CertData c) {
		encoding_length++;
		compSizeByteArray(c.Name);
		compSize(c.PublicKey);
	}

	private void encodeDate(Date c) {
		encodeDestination[encoding_length++] = Code.DATE;
		encodeShort((short) c.year);
		encodeShort((short) c.dayOfYear);
	}

	private void decodeDateInto(byte[] in, Date into) {
		if (in[encoding_length++] != Code.DATE)
			SimpleComm.raiseError();
		into.year = decodeShort(in);
		into.dayOfYear = decodeShort(in);
	}

	private void compSizeDate(Date c) {
		encoding_length++;
		compSizeShort(c.year);
		compSizeShort(c.dayOfYear);
	}

	private void encodePublicKey(PublicKey c) {
		encodeDestination[encoding_length++] = Code.PUBLICKEY;

		encodeByteArray(c.modulus);
	}

	private void decodePublicKeyInto(byte[] in, PublicKey into) {
		if (in[encoding_length++] != Code.PUBLICKEY)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.modulus);
		into.updateKey();
	}

	private void compSizePublicKey(PublicKey c) {
		encoding_length++;

		compSizeByteArray(c.modulus);

	}

	public void encodePrivateKey(PrivateKey c) {
		encodeDestination[encoding_length++] = Code.PRIVATEKEY;

		encodeByteArray(c.modulus);
		encodeByteArray(c.privexponent);

	}

	private void decodePrivateKeyInto(byte[] in, PrivateKey into) {
		if (in[encoding_length++] != Code.PRIVATEKEY)
			SimpleComm.raiseError();
		decodeByteArrayInto(in, into.modulus);
		decodeByteArrayInto(in, into.privexponent);
		into.updateKey();
	}

	private void compSizePrivateKey(PrivateKey c) {
		encoding_length++;

		compSizeByteArray(c.modulus);
		compSizeByteArray(c.privexponent);

	}

	public void decodeEPa(byte[] in, short offset, EPa thisEPa) {
		encoding_length = offset;
		if (in[encoding_length++] != Code.EPA)
			SimpleComm.raiseError();

		if (in[encoding_length] != Code.IGNORE) {
			decodeCertificateInto(in, thisEPa.EPaCert);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodeByteArrayInto(in, thisEPa.Name);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			thisEPa.State = decodeByte(in);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodeByteArrayInto(in, thisEPa.answer);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodeDateInto(in, thisEPa.dateOfBirth);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodePrivateKeyInto(in, thisEPa.ownPrivkey);
		} else
			encoding_length++;

		if (in[encoding_length] != Code.IGNORE) {
			decodePublicKeyInto(in, thisEPa.trustedPukey);
		} else
			encoding_length++;
	}

}
