package generated.ageVerification.terminal.terminal;

import swt.util.ByteArray;
import java.nio.charset.Charset;

/**
 * Class for T(L)V encoding of messages and other classes
 * 
 */
public class Coding {

	/**
	 * offset into encodeDestination for writing/reading
	 */
	private short encoding_length;

	/**
	 * array used for encoding/decoding
	 */
	private byte[] encodeDestination;

	/**
	 * Singleton instance
	 */
	private static Coding the_instance;

	private static Charset charset = Charset.forName("UTF-8");

	/**
	 * Singleton accessor
	 * 
	 * @return Singleton instance
	 */

	public static Coding getInstance() {
		if (the_instance == null)
			the_instance = new Coding();
		return the_instance;
	}
	/**
	 * return offset into encodeDestination
	 * 
	 * @return value of encoding_length
	 */
	public short getEncodingLength() {
		return encoding_length;
	}

	/**
	 * initializes encoding. This method must be called before encoding an object.
	 * The offset (encoding_length) is reset to 0 and encodeDestination is set to dest.
	 * 
	 * @param dest The (empty) destination array for encoding
	 */
	public void encodeInit(byte[] dest) {
		encodeDestination = dest;
		encoding_length = 0;
	}
	/**
	 * initializes decoding. The offset (encoding_length) is reset to 0.
	 */
	public void decodeInit() {
		encoding_length = 0;
	}

	/**
	 * initializes encoding and encodes an object.
	 * 
	 * @param o The object to encode
	 * @param dest destination array for the encoding
	 * @return length of the encoded object
	 */

	/**
	 * decodes an object beginning at given offset
	 * 
	 * @param in array containing encoded object
	 * @param offset where the encoded object begins
	 * @return the decoded object
	 */

	private byte decodeByte(byte[] in) throws TerminalException {
		if (in[encoding_length++] != Code.BYTE) {

			stop();
		}

		byte res = in[encoding_length++];
		return res;
	}

	private boolean decodeBoolean(byte[] in) throws TerminalException {
		if (in[encoding_length++] != Code.BOOLEAN) {
			stop();

		}
		boolean res = (in[encoding_length++] == 1 ? true : false);
		return res;
	}

	public Codeable decode(byte[] in, short offset) throws Exception {
		encoding_length = offset;
		return decode(in);
	}
	public byte[] stringToByte(String s) {
		byte[] b = s.getBytes(charset);
		byte[] ret = new byte[LengthConstants.LENGTHOFSTRING];
		int len = (b.length < LengthConstants.LENGTHOFSTRING)
				? b.length
				: LengthConstants.LENGTHOFSTRING;
		ByteArray.copy(b, 0, ret, 0, len);
		return ret;
	}

	public boolean isValidCardString(String s) {
		return (s.getBytes(charset).length <= LengthConstants.LENGTHOFSTRING);
	}

	public void encodeString(String i) {
		byte[] o = stringToByte(i);
		encodeDestination[encoding_length++] = Code.BYTEARRAY;
		ByteArray.setShort(encodeDestination, encoding_length,
				(short) (o.length));
		encoding_length += 2;
		ByteArray.copy(o, 0, encodeDestination, encoding_length, o.length);
		encoding_length += o.length;
	}

	private void compSizeString(String i) {
		byte[] o = stringToByte(i);
		encoding_length += 3;
		encoding_length += o.length;
	}

	public short encode(Codeable o, byte[] dest) throws Exception {
		encodeInit(dest);

		encode(o);
		return computeSize(o);

	}

	/**
	 * computes the encoding length of an object without actually encoding it.
	 * 
	 * @param o The object of which the encoding length should be computed
	 * @return encoding length
	 */

	public short computeSize(Codeable o) {
		encoding_length = 0;
		try {
			compSize(o);
			return encoding_length;
		} catch (Exception e) {
			return -1;
		}
	}

	public void encodeByte(byte o) {
		encodeDestination[encoding_length++] = Code.BYTE;
		encodeDestination[encoding_length++] = o;
	}

	private void encodeBoolean(boolean o) {
		encodeDestination[encoding_length++] = Code.BOOLEAN;
		encodeDestination[encoding_length++] = (byte) (o ? 1 : 0);
	}

	private void compSizeByte(byte o) {
		encoding_length += 2;
	}

	public void encodeInt(int i) {
		short o = (short) i;
		if (i != (int) o)
			System.err.println("int too big for cast into short.");
		encodeDestination[encoding_length++] = Code.SHORT;
		ByteArray.setShort(encodeDestination, encoding_length, o);
		encoding_length += 2;
	}

	private void compSizeInt(int i) {
		short o = (short) i;
		encoding_length += 3;
	}
	public int decodeInt(byte[] in) {
		short res = ByteArray.getShort(in, ++encoding_length);
		encoding_length += 2;
		return (int) res;
	}

	public byte[] decodeByteArray(byte[] in) {
		short len = ByteArray.getShort(in, ++encoding_length);
		byte[] dest = new byte[len];
		encoding_length += 2;
		ByteArray.copy(in, encoding_length, dest, 0, len);
		encoding_length += len;
		return dest;
	}

	public String byteToString(byte[] b) {
		int len = b.length;
		while (len > 0 && b[len - 1] == 0)
			len--;
		return new String(ByteArray.subarray(b, 0, len), charset);
	}

	public String decodeString(byte[] in) {
		short len = ByteArray.getShort(in, ++encoding_length);
		byte[] dest = new byte[len];
		encoding_length += 2;
		ByteArray.copy(in, encoding_length, dest, 0, len);
		encoding_length += len;
		String str = byteToString(dest);
		return str;
	}

	private void compSizeBoolean(boolean o) {
		encoding_length += 2;
	}

	public byte[] encodeSingleString(String s) {
		byte[] o = stringToByte(s);
		byte[] dest = new byte[3 + o.length];
		int i = 0;
		dest[i++] = Code.BYTEARRAY;
		ByteArray.setShort(dest, i, (short) (o.length));
		i += 2;
		ByteArray.copy(o, 0, dest, i, o.length);
		i += o.length;
		return dest;
	}

	public void encodeByteArray(byte[] o) {

		//byte[] o = stringToByte(i);

		encodeDestination[encoding_length++] = Code.BYTEARRAY;

		ByteArray.setShort(encodeDestination, encoding_length,
				(short) (o.length));
		encoding_length += 2;

		ByteArray.copy(o, 0, encodeDestination, encoding_length, o.length);

		encoding_length += o.length;
	}

	private void compSizeByteArray(byte[] o) {
		encoding_length += 3;
		encoding_length += o.length;
	}

	/**
	 * encodes an object without initializing destination and offset.
	 * 
	 * @param c object to encode
	 */

	public void encode(Codeable c) throws Exception {

		switch (c.getCode()) {
			case Code.SENDCERT :
				encodeSendCert((sendCert) c);
				break;
			case Code.SENDNONCE :
				encodeSendNonce((sendNonce) c);
				break;
			case Code.SIGBACK :
				encodeSigBack((sigBack) c);
				break;
			case Code.SIG :
				encodeSig((sig) c);
				break;
			case Code.USERINSERTEPA :
				encodeUserInsertEPa((UserInsertEPa) c);
				break;
			case Code.USEROK :
				encodeUserOK((UserOK) c);
				break;
			case Code.NONCE :
				encodeNonce((Nonce) c);
				break;
			case Code.CERTIFICATE :
				encodeCertificate((Certificate) c);
				break;
			case Code.SIGBACKDATA :
				encodeSigBackData((SigBackData) c);
				break;
			case Code.SIGNEDDATA :
				encodeSignedData((SignedData) c);
				break;
			case Code.SIGDATA :
				encodeSigData((SigData) c);
				break;
			case Code.CERTDATA :
				encodeCertData((CertData) c);
				break;
			case Code.DATE :
				encodeDate((Date) c);
				break;
			case Code.PUBLICKEY :
				encodePublicKey((PublicKey) c);
				break;
			case Code.PRIVATEKEY :
				encodePrivateKey((PrivateKey) c);
				break;
			// never reached
			default :

				stop();

		}
	}

	public Codeable decode(byte[] in) throws Exception {
		switch (in[encoding_length]) {
			case Code.SENDCERT :
				return decodeSendCert(in);
			case Code.SENDNONCE :
				return decodeSendNonce(in);
			case Code.SIGBACK :
				return decodeSigBack(in);
			case Code.SIG :
				return decodeSig(in);
			case Code.USERINSERTEPA :
				return decodeUserInsertEPa(in);
			case Code.USEROK :
				return decodeUserOK(in);
			case Code.NONCE :
				return decodeNonce(in);
			case Code.CERTIFICATE :
				return decodeCertificate(in);
			case Code.SIGBACKDATA :
				return decodeSigBackData(in);
			case Code.SIGNEDDATA :
				return decodeSignedData(in);
			case Code.SIGDATA :
				return decodeSigData(in);
			case Code.CERTDATA :
				return decodeCertData(in);
			case Code.DATE :
				return decodeDate(in);
			case Code.PUBLICKEY :
				return decodePublicKey(in);
			case Code.PRIVATEKEY :
				return decodePrivateKey(in);
				// never reached
			default :
				stop();
		}
		return null;
	}

	public void compSize(Codeable c) throws Exception {
		switch (c.getCode()) {
			case Code.SENDCERT :

				compSizeSendCert((sendCert) c);

				break;
			case Code.SENDNONCE :

				compSizeSendNonce((sendNonce) c);

				break;
			case Code.SIGBACK :

				compSizeSigBack((sigBack) c);

				break;
			case Code.SIG :

				compSizeSig((sig) c);

				break;
			case Code.USERINSERTEPA :

				compSizeUserInsertEPa((UserInsertEPa) c);

				break;
			case Code.USEROK :

				compSizeUserOK((UserOK) c);

				break;
			case Code.NONCE :

				compSizeNonce((Nonce) c);

				break;
			case Code.CERTIFICATE :

				compSizeCertificate((Certificate) c);

				break;
			case Code.SIGBACKDATA :

				compSizeSigBackData((SigBackData) c);

				break;
			case Code.SIGNEDDATA :

				compSizeSignedData((SignedData) c);

				break;
			case Code.SIGDATA :

				compSizeSigData((SigData) c);

				break;
			case Code.CERTDATA :

				compSizeCertData((CertData) c);

				break;
			case Code.DATE :

				compSizeDate((Date) c);

				break;
			case Code.PUBLICKEY :

				compSizePublicKey((PublicKey) c);

				break;
			case Code.PRIVATEKEY :

				compSizePrivateKey((PrivateKey) c);

				break;
			// never reached
			default :

				stop();

		}
	}

	public void encodeSendCert(sendCert c) throws Exception {
		encodeDestination[encoding_length++] = Code.SENDCERT;

		encode(c.getNonce1());

		encode(c.getTCert());
	}

	public sendCert decodeSendCert(byte[] in) throws Exception {
		sendCert c = new sendCert();
		encoding_length++;
		c.setNonce1(((Nonce) decode(in)));
		c.setTCert(((Certificate) decode(in)));
		return c;
	}

	private void compSizeSendCert(sendCert c) throws Exception {
		encoding_length++;
		compSize(c.getNonce1());
		compSize(c.getTCert());
	}

	public void encodeSendNonce(sendNonce c) throws Exception {
		encodeDestination[encoding_length++] = Code.SENDNONCE;

		encode(c.getNonce2());
	}

	public sendNonce decodeSendNonce(byte[] in) throws Exception {
		sendNonce c = new sendNonce();
		encoding_length++;
		c.setNonce2(((Nonce) decode(in)));
		return c;
	}

	private void compSizeSendNonce(sendNonce c) throws Exception {
		encoding_length++;
		compSize(c.getNonce2());
	}

	public void encodeSigBack(sigBack c) throws Exception {
		encodeDestination[encoding_length++] = Code.SIGBACK;

		encode(c.getData());

		encode(c.getSig());

		encode(c.getEPaCert());
	}

	public sigBack decodeSigBack(byte[] in) throws Exception {
		sigBack c = new sigBack();
		encoding_length++;
		c.setData(((SigBackData) decode(in)));
		c.setSig(((SignedData) decode(in)));
		c.setEPaCert(((Certificate) decode(in)));
		return c;
	}

	private void compSizeSigBack(sigBack c) throws Exception {
		encoding_length++;
		compSize(c.getData());
		compSize(c.getSig());
		compSize(c.getEPaCert());
	}

	public void encodeSig(sig c) throws Exception {
		encodeDestination[encoding_length++] = Code.SIG;

		encode(c.getData());

		encode(c.getSig());
	}

	public sig decodeSig(byte[] in) throws Exception {
		sig c = new sig();
		encoding_length++;
		c.setData(((SigData) decode(in)));
		c.setSig(((SignedData) decode(in)));
		return c;
	}

	private void compSizeSig(sig c) throws Exception {
		encoding_length++;
		compSize(c.getData());
		compSize(c.getSig());
	}

	public void encodeUserInsertEPa(UserInsertEPa c) throws Exception {
		encodeDestination[encoding_length++] = Code.USERINSERTEPA;

		encodeInt((int) c.getAge());
	}

	public UserInsertEPa decodeUserInsertEPa(byte[] in) throws Exception {
		UserInsertEPa c = new UserInsertEPa();
		encoding_length++;
		c.setAge(decodeInt(in));
		return c;
	}

	private void compSizeUserInsertEPa(UserInsertEPa c) throws Exception {
		encoding_length++;

		compSizeInt(c.getAge());
	}

	public void encodeUserOK(UserOK c) throws Exception {
		encodeDestination[encoding_length++] = Code.USEROK;
	}

	public UserOK decodeUserOK(byte[] in) throws Exception {
		UserOK c = new UserOK();
		encoding_length++;
		return c;
	}

	private void compSizeUserOK(UserOK c) throws Exception {
		encoding_length++;
	}

	public void encodeNonce(Nonce c) throws Exception {
		encodeDestination[encoding_length++] = Code.NONCE;

		encodeByteArray((byte[]) c.getNonce());
	}

	public Nonce decodeNonce(byte[] in) throws Exception {
		Nonce c = new Nonce();
		encoding_length++;
		c.setNonce(decodeByteArray(in));
		return c;
	}

	private void compSizeNonce(Nonce c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getNonce());
	}

	public void encodeCertificate(Certificate c) throws Exception {
		encodeDestination[encoding_length++] = Code.CERTIFICATE;

		encode(c.getData());

		encode(c.getSig());
	}

	public Certificate decodeCertificate(byte[] in) throws Exception {
		Certificate c = new Certificate();
		encoding_length++;
		c.setData(((CertData) decode(in)));
		c.setSig(((SignedData) decode(in)));
		return c;
	}

	private void compSizeCertificate(Certificate c) throws Exception {
		encoding_length++;
		compSize(c.getData());
		compSize(c.getSig());
	}

	public void encodeSigBackData(SigBackData c) throws Exception {
		encodeDestination[encoding_length++] = Code.SIGBACKDATA;

		encode(c.getNonce1());

		encode(c.getNonce2());

		encodeInt((int) c.getAge());

		encodeString((String) c.getAnswer());
	}

	public SigBackData decodeSigBackData(byte[] in) throws Exception {
		SigBackData c = new SigBackData();
		encoding_length++;
		c.setNonce1(((Nonce) decode(in)));
		c.setNonce2(((Nonce) decode(in)));
		c.setAge(decodeInt(in));
		c.setAnswer(decodeString(in));
		return c;
	}

	private void compSizeSigBackData(SigBackData c) throws Exception {
		encoding_length++;
		compSize(c.getNonce1());
		compSize(c.getNonce2());

		compSizeInt(c.getAge());

		compSizeString(c.getAnswer());
	}

	public void encodeSignedData(SignedData c) throws Exception {
		encodeDestination[encoding_length++] = Code.SIGNEDDATA;

		encodeByteArray((byte[]) c.getSigned());
	}

	public SignedData decodeSignedData(byte[] in) throws Exception {
		SignedData c = new SignedData();
		encoding_length++;
		c.setSigned(decodeByteArray(in));
		return c;
	}

	private void compSizeSignedData(SignedData c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getSigned());
	}

	public void encodeSigData(SigData c) throws Exception {
		encodeDestination[encoding_length++] = Code.SIGDATA;

		encode(c.getNonce1());

		encode(c.getNonce2());

		encodeInt((int) c.getAge());

		encode(c.getDatum());
	}

	public SigData decodeSigData(byte[] in) throws Exception {
		SigData c = new SigData();
		encoding_length++;
		c.setNonce1(((Nonce) decode(in)));
		c.setNonce2(((Nonce) decode(in)));
		c.setAge(decodeInt(in));
		c.setDatum(((Date) decode(in)));
		return c;
	}

	private void compSizeSigData(SigData c) throws Exception {
		encoding_length++;
		compSize(c.getNonce1());
		compSize(c.getNonce2());

		compSizeInt(c.getAge());
		compSize(c.getDatum());
	}

	public void encodeCertData(CertData c) throws Exception {
		encodeDestination[encoding_length++] = Code.CERTDATA;

		encodeString((String) c.getName());

		encode(c.getPublicKey());
	}

	public CertData decodeCertData(byte[] in) throws Exception {
		CertData c = new CertData();
		encoding_length++;
		c.setName(decodeString(in));
		c.setPublicKey(((PublicKey) decode(in)));
		return c;
	}

	private void compSizeCertData(CertData c) throws Exception {
		encoding_length++;

		compSizeString(c.getName());
		compSize(c.getPublicKey());
	}

	public void encodeDate(Date c) throws Exception {
		encodeDestination[encoding_length++] = Code.DATE;

		encodeInt((int) c.getYear());

		encodeInt((int) c.getDayOfYear());
	}

	public Date decodeDate(byte[] in) throws Exception {
		Date c = new Date();
		encoding_length++;
		c.setYear(decodeInt(in));
		c.setDayOfYear(decodeInt(in));
		return c;
	}

	private void compSizeDate(Date c) throws Exception {
		encoding_length++;

		compSizeInt(c.getYear());

		compSizeInt(c.getDayOfYear());
	}

	private void encodePublicKey(PublicKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.PUBLICKEY;
		encodeByteArray(c.getModulus());

	}

	public PublicKey decodePublicKey(byte[] in) throws Exception {
		if (in[encoding_length++] != Code.PUBLICKEY)
			stop();
		PublicKey res = new PublicKey();
		res.setKey(decodeByteArray(in));
		return res;
	}

	private void compSizePublicKey(PublicKey c) throws Exception {
		encoding_length++;
		compSizeByteArray(c.getModulus());

	}

	public void encodePrivateKey(PrivateKey c) throws Exception {
		encodeDestination[encoding_length++] = Code.PRIVATEKEY;

		encodeByteArray(c.getModulus());
		encodeByteArray(c.getPrivexponent());

	}

	public PrivateKey decodePrivateKey(byte[] in) throws Exception {
		if (in[encoding_length++] != Code.PRIVATEKEY)
			stop();
		PrivateKey res = new PrivateKey();
		res.setKey(decodeByteArray(in), decodeByteArray(in));
		return res;
	}

	private void compSizePrivateKey(PrivateKey c) throws Exception {
		encoding_length++;

		compSizeByteArray(c.getModulus());
		compSizeByteArray(c.getPrivexponent());

	}

	private void stop() throws TerminalException {
		throw new TerminalException("Coding");
	}
}
