package generated.ageVerification.terminal.terminal;

//import swt.simulation.SmartCardSimulation;
import swt.terminal.HandleResult;
import swt.util.ByteArray;
import swt.util.HexString;
import de.uniaugsburg.swt.javacard.terminalinterface.SWTReader;
import de.uniaugsburg.swt.javacard.terminalinterface.SWTReaderException;
import javax.xml.ws.BindingProvider;

import once.ageVerification.terminal.*;

public class Terminal {

	private PublicKey trustedPukey;

	private PrivateKey ownPrivkey;

	private String expectedAnswer;

	private Nonce challenge;

	private Nonce cardchallenge;

	private int age;

	private Certificate CertificateT;

	private Userinterface userinterface;
	public void setUserinterface(Userinterface u) {
		this.userinterface = u;
	}

	// saved prev message
	public byte[] prevMsg;
	public UserMessage prevUMsg;

	public final Gate dummyGate = new Gate() {
		@Override
		public void callConstraints() {
			/* do nothing */
		}

		@Override
		public byte[] transmit(Terminal terminal, byte[] apdu, int port)
				throws TerminalException, SWTReaderException {
			return terminal.transmit(apdu, port);
		}

		@Override
		public void finishProtocol(Terminal terminal) {
			/* do nothing */
		}

	};
	public Gate currentGate = dummyGate;

	/**
	 * Resets the current gate to a default dummy gate
	 */
	public void resetGate() {
		currentGate = dummyGate;
	}

	/**
	 * Sets the current gate
	 */
	public void setGate(Gate gate) {
		currentGate = gate;
	}

	public Terminal() throws java.lang.Exception {

		trustedPukey = new PublicKey();
		ownPrivkey = new PrivateKey();
		expectedAnswer = new String();
		challenge = new Nonce();
		cardchallenge = new Nonce();
		age = 0;
		CertificateT = new Certificate();

	}

	public Terminal(Certificate CertificateT, String expectedAnswer,
			PrivateKey ownPrivkey, PublicKey trustedPukey)
			throws java.lang.Exception {

		this.CertificateT = CertificateT;
		this.expectedAnswer = expectedAnswer;
		this.ownPrivkey = ownPrivkey;
		this.trustedPukey = trustedPukey;

		challenge = new Nonce();
		cardchallenge = new Nonce();
		age = 0;

	}

	public Terminal(Certificate CertificateT, byte[] expectedAnswer,
			PrivateKey ownPrivkey, PublicKey trustedPukey) {
		this.CertificateT = CertificateT;
		this.expectedAnswer = Coding.getInstance().byteToString(expectedAnswer);
		this.ownPrivkey = ownPrivkey;
		this.trustedPukey = trustedPukey;

		challenge = new Nonce();
		cardchallenge = new Nonce();
		age = 0;

	}

	private void sendMsg(Message msg) {

		currentGate.callConstraints();
		userinterface.send((UserMessage) msg);

	}

	private void sendMsg(Message msg, int port) throws java.lang.Exception {

		sendMsg(msg, port, false, false);

	}
	private void sendMsg(Message msg, int port, Boolean openSessionBeforeSend,
			Boolean closeSessionAfterReceive) throws java.lang.Exception {

		currentGate.callConstraints();

		Message response;

		switch (port) {

			case Ports.Terminal2EPa_default :
				response = sendCardMsg(msg, port);
				if (response != null)
					processMessage(response);
				break;

			default :
				stop();
		}
	}

	private Message sendCardMsg(Message msg, int port)
			throws java.lang.Exception {
		byte[] encMsg = new byte[LengthConstants.MAX_ENCODING_LENGTH_MESSAGES];
		short encMsgLen = Coding.getInstance().encode((Codeable) msg, encMsg);
		encMsg = ByteArray.subarray(encMsg, 0, encMsgLen);
		byte[] apdu = ByteArray.append(new byte[]{0, 2, 0, 0, 0}, encMsg);
		byte[] response = currentGate.transmit(this, apdu, port);

		if (response.length == 2) {
			if (!(response[0] == (byte) 0x90 && response[1] == (byte) 0x00))
				throw new TerminalException("Cardexception: "
						+ HandleResult.interpretSW(response));
			return null;
		} else {
			prevMsg = response;

			Coding.getInstance().decodeInit();
			Codeable cResponse = Coding.getInstance().decode(response);
			return (Message) cResponse;
		}
	}

	/**
	 * Processes a raw data stream by trying to decode data to Message.
	 * Needed for testcases.
	 * @param data
	 */
	public void processRawMessage(byte[] data) {
		prevMsg = data;
		try {
			Coding.getInstance().decodeInit();
			Codeable cResponse = Coding.getInstance().decode(data);
			processMessage((Message) cResponse);
		} catch (java.lang.Exception e) {
			System.err.println("Failed decoding data in processRawMessage");
		}

	}

	private void processMessage(Message inmsg) throws java.lang.Exception {
		currentGate.callConstraints();

		switch (inmsg.getCode()) {
			case Code.USERINSERTEPA :
				processUserInsertEPa((UserInsertEPa) inmsg);
				break;
			case Code.SENDNONCE :
				processSendNonce((sendNonce) inmsg);
				break;
			case Code.SIGBACK :
				processSigBack((sigBack) inmsg);
				break;
			default :
				stop();
		}

		currentGate.callConstraints();
		currentGate.finishProtocol(this);
	}

	public void process(Message msg) throws TerminalException {
		currentGate.callConstraints();

		try {
			if (msg instanceof UserMessage)
				prevUMsg = (UserMessage) msg;
			processMessage(msg);
		} catch (java.lang.Exception e) {
			// e.printStackTrace();
			if (e instanceof TerminalException)
				throw (TerminalException) e;
			else
				throw new TerminalException(
						"Error when calling process method: " + e);
		}

		currentGate.callConstraints();
		currentGate.finishProtocol(this);
	}

	private final static int MAX_SPLIT_LENGTH = 50;
	private final static int MAX_APDU_LENGTH = 254;
	private final static int APDU_HEADER_LENGTH = 5;
	private final static int APDU_DATA_LENTH = MAX_APDU_LENGTH
			- APDU_HEADER_LENGTH;
	private SWTReader[] reader = new SWTReader[1];

	public void initReader(SWTReader reader, int port) {
		this.reader[port] = reader;
	}

	public byte[] encode(Codeable c) throws java.lang.Exception {
		byte[] encMsg = new byte[LengthConstants.MAX_ENCODING_LENGTH_MESSAGES];
		short encMsgLen = Coding.getInstance().encode(c, encMsg);
		return ByteArray.subarray(encMsg, 0, encMsgLen);
	}

	public byte[] transmit(byte[] apdu, int port) throws TerminalException,
			SWTReaderException {
		byte[] received = null;
		if (apdu.length <= MAX_APDU_LENGTH) {
			apdu[4] = (byte) (apdu.length - 5);
			received = reader[port].transmit(apdu);
			if (received.length == 2)
				return received;
			received = ByteArray.subarray(received, 0, received.length - 2);
		} else {
			byte[] splittedApdu;
			byte[] apduHeader = ByteArray.subarray(apdu, 0, APDU_HEADER_LENGTH);
			apdu = ByteArray.subarray(apdu, APDU_HEADER_LENGTH);
			int splitCount = apdu.length / APDU_DATA_LENTH;
			if (apdu.length % APDU_DATA_LENTH != 0)
				splitCount++;
			for (int i = 0; i < splitCount; i++) {
				if (i == splitCount - 1) {
					splittedApdu = ByteArray
							.subarray(apdu, i * APDU_DATA_LENTH);
					apduHeader[2] = (byte) 0;
					apduHeader[4] = (byte) splittedApdu.length;
					splittedApdu = ByteArray.append(apduHeader, splittedApdu);
					received = reader[port].transmit(splittedApdu);
					if (received.length == 2)
						return received;
					received = ByteArray.subarray(received, 0,
							received.length - 2);
				} else {
					splittedApdu = ByteArray.subarray(apdu,
							i * APDU_DATA_LENTH, APDU_DATA_LENTH);
					apduHeader[2] = (byte) 1;
					apduHeader[4] = (byte) splittedApdu.length;
					splittedApdu = ByteArray.append(apduHeader, splittedApdu);
					received = reader[port].transmit(splittedApdu);
					if (!(received[0] == (byte) 0x90 && received[1] == (byte) 0x00)) {
						throw new TerminalException(
								"Bad answer from card: part of a long APDU was not answered with OK (0x9000).");
					}
				}
			}
		}

		for (int i = 0; i < MAX_SPLIT_LENGTH; i++) {
			if (received[0] == 0) {
				received = ByteArray.subarray(received, 1);
				return received;
			} else if (received[0] == 1) {
				received = ByteArray.subarray(received, 1);
				byte[] getMore = new byte[]{0, 6, 0, 0, 0, 1, 1};
				byte[] temp_received = reader[port].transmit(getMore);

				temp_received = ByteArray.subarray(temp_received, 0,
						temp_received.length - 2);
				received = ByteArray.append(temp_received[0], received);
				temp_received = ByteArray.subarray(temp_received, 1);
				received = ByteArray.append(received, temp_received);
			} else
				throw new TerminalException(
						"Bad answer from card: answer to a long APDU must begin with 0 or 1.");
		}
		return received;
	}

	/**
	 * This method initializes the card with the given parameters that correspond 
	 * to the card attributes. A null value will be ignored by the card.
	 * @throws Exception
	 */
	public void initCard_EPa(int port, Certificate EPaCert, String Name,
			byte State, String answer, Date dateOfBirth, PrivateKey ownPrivkey,
			PublicKey trustedPukey) throws java.lang.Exception {
		byte cardCode = Code.EPA;
		byte[] apdu = {0, 4, 0, 0, 0, cardCode};
		apdu = ByteArray.append(apdu, (EPaCert != null)
				? encode(EPaCert)
				: new byte[]{Code.IGNORE});

		if (!Coding.getInstance().isValidCardString(Name))
			System.err.println("Encoded string " + Name + " is longer than "
					+ LengthConstants.LENGTHOFSTRING + " bytes.");
		apdu = ByteArray.append(apdu, (Name != null) ? Coding.getInstance()
				.encodeSingleString(Name) : new byte[]{Code.IGNORE});
		apdu = ByteArray.append(apdu, new byte[]{Code.BYTE, State});
		if (!Coding.getInstance().isValidCardString(answer))
			System.err.println("Encoded string " + answer + " is longer than "
					+ LengthConstants.LENGTHOFSTRING + " bytes.");
		apdu = ByteArray.append(apdu, (answer != null) ? Coding.getInstance()
				.encodeSingleString(answer) : new byte[]{Code.IGNORE});
		apdu = ByteArray.append(apdu, (dateOfBirth != null)
				? encode(dateOfBirth)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (ownPrivkey != null)
				? encode(ownPrivkey)
				: new byte[]{Code.IGNORE});

		apdu = ByteArray.append(apdu, (trustedPukey != null)
				? encode(trustedPukey)
				: new byte[]{Code.IGNORE});

		transmit(apdu, port);
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processUserInsertEPa(UserInsertEPa inmsg)
			throws java.lang.Exception {
		synchronized (this) {
			int ageToCheck = inmsg.getAge();
			setAge(ageToCheck);
			challenge.copy(Nonce.generateNonce());
			sendMsg(new sendCert(challenge, CertificateT),
					Ports.Terminal2EPa_default);
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processSendNonce(sendNonce inmsg) throws java.lang.Exception {
		synchronized (this) {
			Nonce no = inmsg.getNonce2();
			cardchallenge.copy(no);
			SigData temp = new SigData(challenge, no, age,
					Calendar.currentDate());
			SignedData signedData = SignedData.sign(ownPrivkey, temp);
			sendMsg(new sig(temp, signedData), Ports.Terminal2EPa_default);
		}
	}

	/**
	 * This method is called, when a -message arrives.
	 * Parameters are a result of the activity diagrams
	 */
	private void processSigBack(sigBack inmsg) throws java.lang.Exception {
		synchronized (this) {
			SigBackData tp = inmsg.getData();
			SignedData signedData = inmsg.getSig();
			Certificate ec = inmsg.getEPaCert();
			boolean valid1 = SignedData.verify(trustedPukey, ec.getSig(),
					ec.getData());
			if (valid1 == true) {
				boolean valid2 = SignedData.verify(ec.getData().getPublicKey(),
						signedData, tp);
				if (valid2 == true) {
					boolean valid3 = tp.equals(new SigBackData(challenge,
							cardchallenge, age, expectedAnswer));
					if (valid3 == true) {
						sendMsg(new UserOK());
					} else {
						stop();
					}
				} else {
					stop();
				}
			} else {
				stop();
			}
		}
	}

	public PublicKey getTrustedPukey() {
		return trustedPukey;
	}

	public PrivateKey getOwnPrivkey() {
		return ownPrivkey;
	}

	public String getExpectedAnswer() {
		return expectedAnswer;
	}

	public Nonce getChallenge() {
		return challenge;
	}

	public Nonce getCardchallenge() {
		return cardchallenge;
	}

	public int getAge() {
		return age;
	}

	public Certificate getCertificateT() {
		return CertificateT;
	}

	public void setTrustedPukey(PublicKey trustedPukey) {
		this.trustedPukey = trustedPukey;
	}
	public void setOwnPrivkey(PrivateKey ownPrivkey) {
		this.ownPrivkey = ownPrivkey;
	}
	public void setExpectedAnswer(String expectedAnswer) {

		if (!Coding.getInstance().isValidCardString(expectedAnswer))
			System.err.println("Encoded string " + expectedAnswer
					+ " is longer than " + LengthConstants.LENGTHOFSTRING
					+ " bytes.");
		this.expectedAnswer = expectedAnswer;
	}
	public void setExpectedAnswer(byte[] expectedAnswer) {
		setExpectedAnswer(Coding.getInstance().byteToString(expectedAnswer));
	}

	public void setChallenge(Nonce challenge) {
		this.challenge = challenge;
	}
	public void setCardchallenge(Nonce cardchallenge) {
		this.cardchallenge = cardchallenge;
	}
	public void setAge(int age) {
		this.age = age;
	}

	public void setCertificateT(Certificate CertificateT) {
		this.CertificateT = CertificateT;
	}

	protected void stop() throws TerminalException {
		currentGate.callConstraints();
		throw new TerminalException("Terminal");
	}

}
